﻿using System;
using System.Collections.Generic;
using SyntaxTree;
namespace optimusParser
{
/// <summary>
/// построение графа потока управления
/// </summary>
class CFGCreator
{
CFGCreator() { }
// список квадрик
private static List<Quadric> qs = new List<Quadric>();
// граф потока управления
private static BaseBlockGraph cfg = new BaseBlockGraph();
// список квадрик-лидеров
private static List<int> leaders = new List<int>();
// список базовых блоков
private static List<BaseBlock> bb = new List<BaseBlock>();
// формирование списка лидеров
private static List<int> findLeaders()
{
// нахождение лидеров
for (int i = 0; i < qs.Count; i++)
{
// первая квадрика - лидер
if (i == 0) leaders.Add(0);
else if (qs[i].Operation == QuadOperation.Goto)
{
// квадрика, на которую есть переход, - лидер
leaders.Add(int.Parse(qs[i].Result.ToString()));
if (i != qs.Count - 1)
// квадрика, следующая за переходом, - лидер
leaders.Add(i + 1);
}
else if (qs[i].Operation == QuadOperation.If)
// квадрика, на которую есть переход, - лидер
leaders.Add(int.Parse(qs[i + 1].Result.ToString()));
}
// сортировка
leaders.Sort();
// удаление повторяющихся элементов
for (int i = 0; i < leaders.Count - 1; i++)
while (i + 1 < leaders.Count - 1 && leaders[i] == leaders[i + 1])
leaders.RemoveAt(i + 1);
return leaders;
}
// создание списка базовых блоков
private static List<BaseBlock> createBBlocks()
{
// базовый блок - последовательность квадрик от одного лидера до последующего
BBlock tBlock;
int cnt = 0;
for (int i = 0; i < leaders.Count - 1; i++)
{
if (i + 1 == leaders.Count - 1)
tBlock = new BBlock(cnt++.ToString(), // имя
// список квадрик
qs.GetRange(leaders[i + 1], qs.Count - leaders[i + 1]));
else
tBlock = new BBlock(cnt++.ToString(), // имя
// список квадрик
qs.GetRange(leaders[i], leaders[i + 1] - leaders[i]));
cfg.AddBBlock(tBlock);
}
return cfg.ListBBlocks;
}
// построение графа
private static BaseBlockGraph createCFG()
{
// формирование связей (добавление дуг) между базовыми блоками
for (int i = 0; i < bb.Count; i++)
{
// если последняя квадрика i-того ББ - оператор перехода
if (cfg[i.ToString()].Quads[cfg[i.ToString()].Quads.Count - 1].Operation ==
QuadOperation.Goto)
{
// берем эту квадрику
Quadric q = cfg[i.ToString()].Quads[cfg[i.ToString()].Quads.Count - 1];
// и ищем куда переход
for (int j = 0; j < bb.Count; j++)
{
if (i != j)
{
List<Quadric> quadrics = cfg[j.ToString()].Quads;
for (int k = 0; k < quadrics.Count; k++)
// если нашли
if (quadrics[k] == qs[int.Parse(q.Result.ToString())])
{
// добавляем ребро в граф
cfg.AddEdge(cfg[i.ToString()].Name, cfg[j.ToString()].Name);
break;
}
}
}
// если 2-ая квадрика с конца - условный оператор,
// нужно добавить связь со следующим ББ
if (cfg[(i).ToString()].Quads.Count >= 2)
if ((cfg[i.ToString()].Quads[cfg[i.ToString()].Quads.Count - 2].Operation
== QuadOperation.If) && (i + 1 < bb.Count))
cfg.AddEdge(cfg[i.ToString()].Name, cfg[(i + 1).ToString()].Name);
}
// иначе добавляем связь со следующим ББ
else if (i + 1 < bb.Count)
cfg.AddEdge(cfg[i.ToString()].Name, cfg[(i + 1).ToString()].Name);
}
// удаление квадрик перехода
for (int i = 0; i < bb.Count; i++)
if (cfg[i.ToString()].Quads[cfg[i.ToString()].Quads.Count - 1].Operation ==
QuadOperation.Goto)
cfg[i.ToString()].Quads.RemoveAt(cfg[i.ToString()].Quads.Count - 1);
// удаление пустых базовых блоков
foreach (BBlock b in cfg.ListBBlocks)
{
if (cfg[b.Name].Quads.Count == 0)
{
// сохранение связей
for (int i = 0; i < cfg.GetParentNodes(b.Name).Count; ++i)
for (int j = 0; j < cfg.GetChildNodes(b.Name).Count; ++j)
cfg.AddEdge(cfg.GetParentNodes(b.Name)[i].Name,
cfg.GetChildNodes(b.Name)[j].Name);
cfg.RemoveBBlock(b.Name);
}
}
return cfg;
}
public static BaseBlockGraph Create(SyntaxTreeNode root)
{
// получаем список квадрик
qs = QuadricCollector.Collect(root);
// формируем список лидеров
leaders = findLeaders();
// создаем список базовых блоков
bb = createBBlocks();
// строим и возвращаем граф
return createCFG();
}
}
}

