﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DicOfSet = System.Collections.Generic.Dictionary<string, System.Collections.Generic.HashSet<string>>;
namespace optimusParser
{
    public enum direction { forward, backward }
    public enum kindOfCollector { union, intersection }
    public class CommonIterativeAlgorithm
    {
        public static void FindInOut(BBGraph bbGraph, direction dir, kindOfCollector col, HashSet<string> V, HashSet<string> T, DicOfSet Gen, DicOfSet Kill, out DicOfSet IN, out DicOfSet OUT)
        {
            //Так будет обозначаться вход/выход из программы.
            string border = "border";

            IN = new DicOfSet();
            OUT = new DicOfSet();
            if (T == null)
                T = new HashSet<string>();

            if (V != null)
                OUT.Add(border, V);
            else OUT.Add(border, new HashSet<string>());
            foreach (BaseBlock bb in bbGraph.ListBBlocks)
                OUT.Add(bb.Name, T);

            //для каждого базового блока здесь будут храниься предыдущие (следующие при обратном обходе) узлы
            DicOfSet Prev = new Dictionary<string, HashSet<string>>();
            foreach (BaseBlock bb in bbGraph.ListBBlocks)
                Prev.Add(bb.Name, new HashSet<string>());
            //инициализация, в зависимости от направления
            if (dir == direction.forward)
            {
                //заполняем Prev узлами-предками
                foreach (BaseBlock bb in bbGraph.ListBBlocks)
                {
                    HashSet<string> hs = new HashSet<string>();
                    foreach (BaseBlock b in bbGraph.GetParentNodes(bb.Name))
                        hs.Add(b.Name);
                    Prev[bb.Name] = hs;
                }
                Prev[bbGraph.FirstBlocks.Name].Add(border);
            }
            else
            {
                //заполняем Prev узлами-потомками
                foreach (BaseBlock bb in bbGraph.ListBBlocks)
                {
                    HashSet<string> hs = new HashSet<string>();
                    foreach (BaseBlock b in bbGraph.GetChildNodes(bb.Name))
                        hs.Add(b.Name);
                    Prev[bb.Name] = hs;
                }
                Prev[bbGraph.LastBlocks.Name].Add(border);
            }

            //основной цикл алгоритма, вся работа идет как при прямом проходе.
            //если на самом деле проход - обратный, нужно после этого цикла поменять
            //местами IN и OUT. 
            bool SomethingChanged = false;
            do
            {
                //тут будут накапливаться все новые OUT'ы
                DicOfSet OutNew = new Dictionary<string, HashSet<string>>();

                //проходим по всем базовым блокам и переприсваиваем IN'ы и OUT'ы
                foreach (BaseBlock bb in bbGraph.ListBBlocks)
                {
                    //
                    if (Prev[bb.Name].Count > 0)
                    {

                        //оператор сбора
                        HashSet<string> WorkingIn = OUT[Prev[bb.Name].ElementAt<string>(0)];
                        for (int i = 1; i < Prev[bb.Name].Count; ++i)
                            if (col == kindOfCollector.union)
                                WorkingIn = new HashSet<string>(WorkingIn.Union<string>(OUT[Prev[bb.Name].ElementAt<string>(i)]));
                            else
                                WorkingIn = new HashSet<string>(WorkingIn.Intersect<string>(OUT[Prev[bb.Name].ElementAt<string>(i)]));
                        IN[bb.Name] = WorkingIn;
                    }
                    else
                        IN[bb.Name] = new HashSet<string>();

                    //передаточная функция
                    OutNew.Add(bb.Name, DFA_tools.transferFuncB(IN[bb.Name], Gen[bb.Name], Kill[bb.Name]));
                }

                //сравнение старого и нового OUT
                SomethingChanged = false;
                foreach (KeyValuePair<string, HashSet<string>> hs in OUT)
                {
                    if (hs.Key == border)
                        continue;
                    HashSet<string> tmp = OutNew[hs.Key];
                    tmp = new HashSet<string>(tmp.Except<string>(hs.Value));
                    if (tmp.Count > 0)
                    {
                        SomethingChanged = true;
                        break;
                    }
                    tmp = new HashSet<string>(hs.Value.Except<string>(OutNew[hs.Key]));
                    if (tmp.Count > 0)
                    {
                        SomethingChanged = true;
                        break;
                    }
                }

                HashSet<string> in_tmp = OUT[border];
                OUT = OutNew;
                OUT.Add(border, in_tmp);

            } while (SomethingChanged);

            //меняем местами множества при обратном обходе
            if (dir == direction.backward)
            {
                DicOfSet TMP = OUT;
                OUT = IN;
                IN = TMP;
            }
        }
    }
}
