﻿using System;
using System.Collections.Generic;
using System.Text;

using DicOfSet = System.Collections.Generic.Dictionary<string, System.Collections.Generic.HashSet<string>>;
namespace optimusParser
{

    class Extractor
    {
        private Extractor() { }
        public static void ExtractUseDef(BBGraph bbGraph, out DicOfSet use, out DicOfSet def)
        {
            use = new DicOfSet();
            def = new DicOfSet();

            foreach (BaseBlock block in bbGraph.ListBBlocks)
            {

                HashSet<string> useB = new HashSet<string>();
                HashSet<string> defB = new HashSet<string>();

                foreach (Quadric q in block.Quads)
                if(q.Operation != QuadOperation.If)
                {
                    double d;
                    if (q.Operation == QuadOperation.FromIndx)
                    {
                        if (!useB.Contains(q.Result.ToString())) defB.Add(q.Result.ToString());
                        string var = q.Var1 + "[" + q.Var2 + "]";
                        if (!defB.Contains(var)) useB.Add(var);
                    }
                    else if (q.Operation == QuadOperation.ToIndx)
                    {
                        string var = q.Result + "[" + q.Var1 + "]";
                        if (!useB.Contains(var)) defB.Add(var);
                        if (!defB.Contains(q.Var2.ToString()))
                            if (!Double.TryParse(q.Var2.ToString(), out d)) useB.Add(q.Var2.ToString());
                    }
                    else
                    {                        
                        if (!defB.Contains(q.Var1.ToString()))
                            if (!Double.TryParse(q.Var1.ToString(), out d)) useB.Add(q.Var1.ToString());
                        if (!useB.Contains(q.Result.ToString())) defB.Add(q.Result.ToString());
                        if (q.Var2 != null)                                                     //унарные операции
                            if (!defB.Contains(q.Var2.ToString()))
                                if (!Double.TryParse(q.Var2.ToString(), out d)) useB.Add(q.Var2.ToString());
                    }
                }

                use.Add(block.Name, useB);
                def.Add(block.Name, defB);
            }
        }

        private static void BlockEGenKill(BBlock curBlock, DicOfSet eGen, DicOfSet eKill)
        {
            DicOfSet genVars = new DicOfSet();
            HashSet<string> killB = new HashSet<string>();
            foreach (Quadric q in curBlock.Quads)
                if (q.Operation != QuadOperation.If)
                {
                    if (q.Operation == QuadOperation.FromIndx)
                    {
                        string eval = q.Var1 + "[" + q.Var2+ "]";
                        if (!genVars.ContainsKey(eval))
                            genVars.Add(eval, new HashSet<string>());
                        genVars[eval].Add(eval);
                        killB.Remove(eval);
                        
                    }
                    else if (q.Operation == QuadOperation.ToIndx)
                    {
                        string eval = q.Result + "[" + q.Var1 + "]";
                        if (genVars.ContainsKey(eval))
                        {
                            killB.Add(eval);
                            genVars.Remove(eval);
                        }
                    }
                    else
                    {
                        if (q.Operation != QuadOperation.Nop)
                        {
                            string eval = q.Var1 + " " + q.Operation + " " + q.Var2;
                            if (!genVars.ContainsKey(q.Var1.ToString()))
                                genVars.Add(q.Var1.ToString(), new HashSet<string>());
                            if(q.Var2!=null)
                            if (!genVars.ContainsKey(q.Var2.ToString()))
                                genVars.Add(q.Var2.ToString(), new HashSet<string>());
                            genVars[q.Var1.ToString()].Add(eval);
                            if (q.Var2 != null)
                            genVars[q.Var2.ToString()].Add(eval);
                            killB.Remove(eval);
                        }
                        if (genVars.ContainsKey(q.Result.ToString()))
                        {
                            killB.UnionWith(genVars[q.Result.ToString()]);
                            HashSet<string> tmp = genVars[q.Result.ToString()];
                            genVars.Remove(q.Result.ToString());
                            foreach (string key in genVars.Keys)
                                genVars[key].ExceptWith(tmp);
                        }
                    }
                }
            eKill.Add(curBlock.Name, killB);
            eGen.Add(curBlock.Name,new HashSet<string>());
            foreach (string s in genVars.Keys)
            {
                eGen[curBlock.Name].UnionWith(genVars[s]);
            }
        }

        public static void ExtractGenKill(BBGraph bbGraph, out DicOfSet gen, out DicOfSet kill)
        {
            gen = new DicOfSet();
            kill = new DicOfSet();
            List<HashSet<string>> t;
            foreach (BBlock b in bbGraph.ListBBlocks)
            {
                t = DFA_tools.getKillD(b, bbGraph);
                gen.Add(b.Name, DFA_tools.getGenB(DFA_tools.getGenD(b), t));
                kill.Add(b.Name, DFA_tools.getKillB(t));
            }

        }

        public static void ExtractEGenKill(BBGraph bbGraph, out DicOfSet eGen, out DicOfSet eKill)
        {
            eGen = new DicOfSet();
            eKill = new DicOfSet();
            foreach(BBlock b in bbGraph.ListBBlocks)
                BlockEGenKill(b as BBlock, eGen, eKill);
        }
    }
}
