%namespace GPPGParserScanner

%using optimusParser;

Alpha [a-zA-Z_]
Digit [0-9]
AlphaDigit {Alpha}|{Digit}
INTNUM {Digit}+
DOUBLENUM {Digit}+\.{Digit}+
ID {Alpha}{AlphaDigit}* 

%%

"=" { return (int)Tokens.ASSIGN; }
";" { return (int)Tokens.SEMICOLUMN; }
"-" { return (int)Tokens.MINUS; }
"+" { return (int)Tokens.PLUS; }
"*" { return (int)Tokens.MULT; }
"/" { return (int)Tokens.DIVIDE; }
"<" { return (int)Tokens.LT; }
">" { return (int)Tokens.GT; }
"<=" { return (int)Tokens.LE; }
">=" { return (int)Tokens.GE; }
"==" { return (int)Tokens.EQ; }
"!=" { return (int)Tokens.NE; }
"(" { return (int)Tokens.LPAREN; }
")" { return (int)Tokens.RPAREN; }
"[" { return (int)Tokens.LSQPAREN; }
"]" { return (int)Tokens.RSQPAREN; }
"," { return (int)Tokens.COLUMN; }
"!" { return (int)Tokens.NOT; }
"||" { return (int)Tokens.OR; }
"&&" { return (int)Tokens.AND; }
"." { return (int)Tokens.COMMA; }
":" { return (int)Tokens.COLON; }
"~" { return (int)Tokens.EXCLAMATION; }
\x01 { return (int)Tokens.INVISIBLE; }

{ID}  { 
  int res = Keywords.KeywordOrIDToken(yytext);
  if (res == (int)Tokens.ID)
    yylval.sVal = yytext;
  return res;
}

{INTNUM} { 
  yylval.iVal = int.Parse(yytext); 
  return (int)Tokens.INTNUM; 
}
{DOUBLENUM} { 
  yylval.dVal = double.Parse(yytext); 
  return (int)Tokens.DOUBLENUM; 
}

%{
  yylloc = new QUT.Gppg.LexLocation(tokLin, tokCol, tokELin, tokECol);
%}

%%

  public override void yyerror(string format, params object[] args) 
  {
    string errorMsg = PT.CreateErrorString(args);
    PT.AddError(errorMsg,yylloc);
  }

//  ,    
public static class Keywords
{
	private static Dictionary<string, int> keywords = new Dictionary<string, int>();

	static Keywords()
	{
		keywords.Add("TRUE", (int)Tokens.TRUE);
		keywords.Add("FALSE", (int)Tokens.FALSE);
		//keywords.Add("ODD", (int)Tokens.ODD);
		//keywords.Add("OR", (int)Tokens.OR);
		keywords.Add("IF", (int)Tokens.IF);
		keywords.Add("THEN", (int)Tokens.THEN);
		keywords.Add("ELSE", (int)Tokens.ELSE);
		keywords.Add("BEGIN", (int)Tokens.BEGIN);
		keywords.Add("END", (int)Tokens.END);
		keywords.Add("WHILE", (int)Tokens.WHILE);
		keywords.Add("DO", (int)Tokens.DO);
		keywords.Add("BOOL", (int)Tokens.BOOL);
		keywords.Add("INT", (int)Tokens.INT);
		keywords.Add("DOUBLE", (int)Tokens.DOUBLE);
		//keywords.Add("CONST", (int)Tokens.CONST);
		//keywords.Add("VAR", (int)Tokens.VAR);
		
	}
	public static int KeywordOrIDToken(string s)
	{
		if (keywords.ContainsKey(s.ToUpper()))
			return keywords[s.ToUpper()];
		else
			return (int)Tokens.ID;
	}
}
  
