%{
//      GPPGParser,   ,   gppg
    public SyntaxTreeNode root; //     
    public GPPGParser(AbstractScanner<ValueType, LexLocation> scanner) : base(scanner) { }
%}

%output=optimus00yacc.cs 

%parsertype GPPGParser

%union  
	{ 
		public bool bVal; 
		public string sVal; 
		public int iVal;
		public double dVal; 
		public Expression ex;
		public VarNode var;
		public VarWithIndex varwi;
		public MainBlock bl;
		public StatementList stl;
		public Statement st;
		public Operation op;		
		public VarList varl;
		public VarWithIndexList varwil;
	}

%using SyntaxTree
%using optimusParser

%namespace GPPGParserScanner

%start MainBlock

%token <sVal> ID
%token <iVal> INTNUM
%token <dVal> DOUBLENUM 
%token <bVal> TRUE FALSE
%token <op> PLUS MINUS MULT DIVIDE AND OR LT GT LE GE EQ NE NOT 
%token ASSIGN SEMICOLUMN LPAREN RPAREN LSQPAREN RSQPAREN COLUMN COMMA COLON EXCLAMATION
%token TRUE FALSE  BOOL INT DOUBLE
%token IF THEN ELSE BEGIN END WHILE DO 
%token INVISIBLE


%type <var> ident
%type <varwi> identwi
%type <ex> expr 
%type <st> Assignment IfStatement WhileStatement  Statement  
%type <st> EmptyStatement  VarDefStatement VarWithIndexDefStatement
%type <stl>   StatementSequence
%type <bl> MainBlock
%type <varl>  VarList
%type <varwil>  VarWithIndexList

%left LT GT LE GE EQ NE
%left PLUS MINUS OR
%left MULT DIVIDE AND 
%left NOT
%left UMINUS

%%

MainBlock
	: StatementSequence
	{
		//     (    )
		root = MainBlock.Create($1);
	}
	;
	
ident 
	: ID {
		$$ = new VarNode($1); 
	}
	;
identwi
	:ID LSQPAREN expr RSQPAREN {
		$$ = new VarWithIndex($1,$3);
	}
	;
	


expr 	
	: ident {
		$$ = $1;
	}
	|identwi{
		$$ = $1;
	}	
	| INTNUM { 
		$$ = new IntConstExpression($1); 		
	}
	| DOUBLENUM { 
		$$ = new DoubleConstExpression($1); 		
	}
	| TRUE {
		$$ = new BoolConstExpression(true);
	}
	| FALSE {
		$$ = new BoolConstExpression(false);
	}
	| MINUS expr %prec UMINUS {
		$$ = new UnaryExpression($2,Operation.minus);
	}
	| LPAREN expr RPAREN {$$ = $2;}
	| NOT expr %prec UNOT {
		$$ = new UnaryExpression($2,Operation.not);
	}
	| expr PLUS expr {
		$$ = new BinaryExpression($1,$3,Operation.plus);
	}
	| expr MINUS expr {
		$$ = new BinaryExpression($1,$3,Operation.minus);
	}
	| expr MULT expr {
		$$ = new BinaryExpression($1,$3,Operation.mult);
	}
	| expr DIVIDE expr {
		$$ = new BinaryExpression($1,$3,Operation.div);
	}
	| expr AND expr {
		$$ = new BinaryExpression($1,$3,Operation.and);
	}
	| expr OR expr {
		$$ = new BinaryExpression($1,$3,Operation.or);
	}
	| expr EQ expr {
		$$ = new BinaryExpression($1,$3,Operation.eq);
	}
	| expr NE expr {
		$$ = new BinaryExpression($1,$3,Operation.ne);
	}
	| expr LT expr {
		$$ = new BinaryExpression($1,$3,Operation.lt);
	}
	| expr LE expr {
		$$ = new BinaryExpression($1,$3,Operation.le);
	}
	| expr GT expr {
		$$ = new BinaryExpression($1,$3,Operation.gt);
	}
	| expr GE expr {
		$$ = new BinaryExpression($1,$3,Operation.ge);
	}
	
	;

Assignment 
	: ident ASSIGN expr {
		$$ = new AssignStatement($1, $3);
	}
	| identwi ASSIGN expr {
		$$ = new AssignStatement($1, $3);
	}
	;

IfStatement	
	: IF expr THEN StatementSequence END {
		$$ = new IfStatement($2, $4, null);
	}
	| IF expr THEN StatementSequence ELSE StatementSequence END {
		$$ = new IfStatement($2, $4, $6);
	}
	
	;

WhileStatement
	: WHILE expr DO StatementSequence END {
		$$ = new WhileStatement($2, $4);
	}
	;

	

EmptyStatement
	: {
		$$ = new EmptyStatement();		
	}
	;
	
Statement 
	: Assignment
	| IfStatement
	| WhileStatement
	| VarDefStatement 
	| VarWithIndexDefStatement
	| EmptyStatement
	;

StatementSequence 
	: Statement {
		$$ = new StatementList($1);
	}
	| StatementSequence SEMICOLUMN Statement {
		$1.Add($3);
		$$ = $1;
	}
	;

VarList 
	: ident {
		$$=new VarList($1);
	}
	| VarList COLUMN ident {
		$1.Add($3);
		$$ = $1;
	}
	;
VarWithIndexList 
	: identwi {
		$$=new VarWithIndexList($1);
	}
	| VarWithIndexList COLUMN identwi {
		$1.Add($3);
		$$ = $1;
	}
	;

VarDefStatement 
	: BOOL VarList  {
	  $$  = new VarDefStatement($2,VarType.Bool);
	}
	| INT VarList  {
	  $$  = new VarDefStatement($2,VarType.Integer);
	}
	|DOUBLE VarList  {
	  $$  = new VarDefStatement($2,VarType.Double);
	}
	;

VarWithIndexDefStatement 
	: BOOL VarWithIndexList  {
	  $$  = new VarWithIndexDefStatement($2,VarType.Bool);
	}
	| INT VarWithIndexList  {
	  $$  = new VarWithIndexDefStatement($2,VarType.Integer);
	}
	|DOUBLE VarWithIndexList  {
	  $$  = new VarWithIndexDefStatement($2,VarType.Double);
	}
	;


	

%%
