﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace optimusParser
{
    /// <summary>Класс содержит в себе методы для локальной оптимизации.</summary>
    class Optimization
    {
        /// <summary>Статический метод - удаляет неиспользуемый код в базовом блоке bblock.</summary>
        /// <param name="bblock">Базовый блок.</param>            
        /// <param name="OUT_LiveVars">Множество переменных живых на выходе из блока.</param>    
        public static void RemoveUnusedCode(BaseBlock bblock, HashSet<string> OUT_LiveVars)
        {
            //множество живых переменных
            HashSet<string> useVars = new HashSet<string>(OUT_LiveVars);            

            //просматриваем квадрики блока в обратном порядке
            for (int i = bblock.CountQuads - 1; i >= 0; i--)
            {
                //добавляем во множество useVars все использования из неудаляемых квадрик
                if (bblock[i].Operation == QuadOperation.If)
                {
                    useVars.Add(bblock[i].Var1.ToString());
                }

                //если во множестве useVars нет текущего генератора, то можно удалить квадрику
                string s = "";
                if (bblock[i].Operation == QuadOperation.ToIndx)
                {
                    s = bblock[i].Result.ToString() + "[" + bblock[i].Var1.ToString() + "]";
                    if (!useVars.Remove(s))
                    {
                        bblock[i] = new Quadric(QuadOperation.Nop, new QVar("", null), new QVar("", null), new QVar("", null));
                        continue;
                    }
                }
                else if (bblock[i].Operation != QuadOperation.Goto && bblock[i].Operation != QuadOperation.If)
                {
                    s = bblock[i].Result.ToString();
                    if (!useVars.Remove(s))
                    {
                        bblock[i] = new Quadric(QuadOperation.Nop, new QVar("", null), new QVar("", null), new QVar("", null));
                        continue;
                    }
                }

                //добавляем во множество useVars все использования из неудаляемых квадрик
                string var1="",var2="",fromInd="";
                if (bblock[i].Var1 != null)
                    var1 = bblock[i].Var1.ToString();
                if (bblock[i].Var2 != null)
                    var2 = bblock[i].Var2.ToString();
                if (bblock[i].Var1 != null && bblock[i].Var2 != null)
                    fromInd = var1 + "[" + var2 + "]";
                if (bblock[i].Operation == QuadOperation.Nop)
                {
                    if (var1 == bblock[i].Result.ToString())
                    {
                        bblock[i] = new Quadric(QuadOperation.Nop, new QVar("", null), new QVar("", null), new QVar("", null));
                        continue;
                    }
                    useVars.Add(var1);
                }
                else if (bblock[i].Operation == QuadOperation.FromIndx)
                {
                    useVars.Add(fromInd);
                    useVars.Add(var2);
                }
                else if (bblock[i].Operation != QuadOperation.Goto && bblock[i].Operation != QuadOperation.If) 
                {                    
                    useVars.Add(var2);
                    useVars.Add(var1);
                }
            }
      
            //удаление отмеченных пустых квадрик
            int k = 0;            
            while(k < bblock.CountQuads)
            {
                if (bblock[k].Result.ToString() == "" && bblock[k].Var1.ToString() == "" && bblock[k].Var2.ToString() == "")
                    bblock.RemoveQuad(k);
                else k++;
            }                     
        }

        /// <summary>
        /// Oптимизирует квадрики базового блока.
        /// </summary>
        /// <param name="bblock">Базовый блок.</param>
        public static void OptimizeQuads(BaseBlock bblock)
        {
            //Для каждой квадрики
           
            foreach (Quadric q in bblock.Quads)
            {
                if (q.Operation == QuadOperation.Plus && (q.Var1.ToString() == "0" || q.Var2.ToString() == "0"))
                {

                    q.Operation = QuadOperation.Nop;
                    if (q.Var1.ToString() == "0")
                    {
                        q.Var1 = q.Var2;
                    }
                    q.Var2 = null;                   
                }
                if (q.Operation == QuadOperation.Mult && (q.Var1.ToString() == "1" || q.Var2.ToString() == "1"))
                {
                    q.Operation = QuadOperation.Nop;
                    if (q.Var1.ToString() == "1")
                    {
                        q.Var1 = q.Var2;
                    }
                    q.Var2 = null;  
                }

                if (q.Var1 is QConst<int> && q.Var2 is QConst<int>)
                {
                    QConst<int> t1 = q.Var1 as QConst<int>;
                    QConst<int> t2 = q.Var2 as QConst<int>;
                    if (q.Operation == QuadOperation.Mult)
                    {
                        q.Var1 = new QConst<int>(t1.Value * t2.Value);
                        q.Var2 = null;
                    }                   
                    if (q.Operation == QuadOperation.Plus)
                    {
                        q.Var1 = new QConst<int>(t1.Value + t2.Value);
                        q.Var2 = null;
                    }
                   
                }

            }

        }
    }
}