﻿using System;
using System.IO;
using System.Text;
using System.Reflection;
using PascalABCCompiler.SyntaxTree;
using System.Collections.Generic;
using optimusParser;
using GPPGParserScanner;

using DicOfSet = System.Collections.Generic.Dictionary<string, System.Collections.Generic.HashSet<string>>;

namespace CreateSyntaxTree
{
    public class MainClass
    {
        public static void Main()
        {
            string FileName = "..\\..\\aaa.opt";
            FileInfo fi = new FileInfo(FileName);
            StreamReader sr = new StreamReader(FileName);
            string Text = sr.ReadToEnd();
            PT.CurrentFileName = FileName;

            Scanner scanner = new Scanner();
            scanner.SetSource(Text, 0);
            //string s=""
            GPPGParser parser = new GPPGParser(scanner);
            //SyntaxTree.         
            if (!parser.Parse())
                if (PT.Errors.Count == 0)
                    PT.AddError("Неопознанная синтаксическая ошибка!", null);
                else
                {
                    for (int i = 0; i < PT.Errors.Count; i++)
                    {
                        Console.WriteLine(PT.Errors[i].Message + " at line " + parser.scanner.yylloc.StartLine + " at column " + parser.scanner.yylloc.StartColumn);
                    }
                }
            else
            {
                BaseBlockGraph g = CFGCreator.Create(parser.root);
                BBGraphVisualizator.CreatePNG(g);
                List<BaseBlock> bl = g.ListBBlocks;
                for (int i = 0; i < bl.Count; i++)
                {
                    for (int k = 0; k < g.GetParentNodes(bl[i].Name).Count; ++k)
                    {
                        Console.WriteLine(i + " <- " + g.GetParentNodes(bl[i].Name)[k].Name);
                    }
                    List<Quadric> quadrics = bl[i].Quads;
                    for (int j = 0; j < quadrics.Count; j++)
                    {
                        Console.WriteLine(String.Format("{0}:\t{1}", i, quadrics[j]));
                    }
                    Console.WriteLine();
                }
                Console.WriteLine("========== Анализ активных переменных ==========");

                DicOfSet IN, OUT;
                DicOfSet Use, Def;
                Extractor.ExtractUseDef(g, out Use, out Def);
                CommonIterativeAlgorithm.FindInOut(g, direction.backward, kindOfCollector.union, null, null, Use, Def, out IN, out OUT);
                for (int i = 0; i < bl.Count; i++)
                {
                    Console.WriteLine("Base Block " + bl[i].Name);
                    Console.Write("IN: ");
                    foreach (string str in IN[bl[i].Name])
                        Console.Write(str + ", ");
                    Console.WriteLine();

                    Console.Write("OUT: ");
                    foreach (string str in OUT[bl[i].Name])
                        Console.Write(str + ", ");
                    Console.WriteLine();
                }

                /////////////////////Удаление неиспользуемого кода (локальное преобразование!)
                /*Console.WriteLine("========== Удаление неиспользуемого кода (локальное преобразование!) ==========");
                for (int i = 0; i < bl.Count; i++)
                    Optimization.RemoveUnusedCode(bl[i], OUT[bl[i].Name]);
                for (int i = 0; i < bl.Count; i++)
                {
                    for (int k = 0; k < g.GetParentNodes(bl[i].Name).Count; ++k)
                    {
                        Console.WriteLine(i + " <- " + g.GetParentNodes(bl[i].Name)[k].Name);
                    }
                    List<Quadric> quadrics = bl[i].Quads;
                    for (int j = 0; j < quadrics.Count; j++)
                    {
                        Console.WriteLine(String.Format("{0}:\t{1}", i, quadrics[j]));
                    }
                    Console.WriteLine();
                }*/
                ////////////////////////////////////////////////////////////////////////////////////

                Console.WriteLine("========== Анализ достигающих определений ==========");
                DicOfSet Kill, Gen;
                Extractor.ExtractGenKill(g, out Gen, out Kill);
                CommonIterativeAlgorithm.FindInOut(g, direction.forward, kindOfCollector.union, null, null, Kill, Gen, out IN, out OUT);
                for (int i = 0; i < bl.Count; i++)
                {
                    Console.WriteLine("Base Block " + bl[i].Name);
                    Console.Write("IN: ");
                    foreach (string str in IN[bl[i].Name])
                        Console.Write(str + ", ");
                    Console.WriteLine();

                    Console.Write("OUT: ");
                    foreach (string str in OUT[bl[i].Name])
                        Console.Write(str + ", ");
                    Console.WriteLine();
                }
                GraphOperations go = new GraphOperations(g);
                go.CyclesFinder();
            }
            Console.ReadLine();
            //parser.root.visit(new Visualizator());
            //parser.root содержит корень синтаксического дерева
            //parser.root
        }

    }
}
