﻿using System;
namespace optimusParser
{
    public enum QuadOperation
    {
        #region SyntaxTreeOp
        /// <summary>
        /// Операции из синтаксического дерева
        /// </summary>

        Plus = 0,
        Minus = 1,
        Mult = 2,
        Div = 3,
        Or = 4,
        And = 5,
        Not = 6,
        Lt = 7,
        Gt = 8,
        Le = 9,
        Ge = 10,
        Eq = 11,
        Ne = 12,
        #endregion

        /// <summary>
        /// Переход.
        /// </summary>
        Goto = 32,
        /// <summary>
        /// Условный переход.
        /// </summary>
        If = 33,
        /// <summary>
        /// Индексная переменная справа.
        /// </summary>
        FromIndx = 34,
        /// <summary>
        /// Индексная переменная слева.
        /// </summary>
        ToIndx = 35,
        /// <summary>
        /// Просто присваивание
        /// </summary>
        Nop = 255
    }

    public class Quadric
    {
        public QuadOperation Operation; // { get { return op; } }
        public QExp Var1; // { get { return v1; } }
        public QExp Var2; // { get { return v2; } }
        public QExp Result; // { get { return res; } }

        //QuadOperation op;
        //string v1;
        //string v2;
        //string res;

        public Quadric(QuadOperation op, QExp v1, QExp v2, QExp res)
        {
            this.Operation = op;
            this.Var1 = v1;
            this.Var2 = v2;
            this.Result = res;
        }

        public override string ToString()
        {            
            switch (Operation)
            {
                case QuadOperation.Goto:
                    return string.Format("goto {0}", Result);
                case QuadOperation.If:
                    return string.Format("if {0} goto {1}", Var1, Result);
                case QuadOperation.Nop:
                    return string.Format("{0} := {1}", Result, Var1);
                case QuadOperation.FromIndx:
                    return string.Format("{0} := {1}[{2}]", Result, Var1, Var2);
                case QuadOperation.ToIndx:
                    return string.Format("{0}[{1}] := {2}", Result, Var1,Var2);
                default:
                    return string.Format("{0} := {1} {2} {3}", Result, Var1, Operation, Var2);
            }
        }
    }

    /// <summary>
    /// Базовый класс выражения для квадрик
    /// </summary>
    public class QExp
    {
        public Type Type;
        public QExp(Type type)
        {
            this.Type = type;
        }
    }

    /// <summary>
    /// Переменная в квадрике
    /// </summary>
    public class QVar:QExp
    {        
        public string Name;
        public QVar(string name, Type type):base(type)
        {            
            this.Name = name;
        }
        public override string ToString()
        {
            return Name;
        }
    }

    /// <summary>
    /// Типизированная константа в квадрике
    /// </summary>
    /// <typeparam name="T">Тип константы</typeparam>
    public class QConst<T>:QExp
    {
        public T Value;
        public QConst(T value):base(typeof(T))
        {
            this.Value = value;            
        }
        public override string ToString()
        {
            return Value.ToString();
        }
    }   
    
}
