﻿using System;
using System.Collections.Generic;
using SyntaxTree;

namespace optimusParser
{
    /// <summary>
    /// Просто пишете QuadricCollector.Collect(parser.root);
    /// и получаете список команд
    /// </summary>
    public class QuadricCollector : Visitor
    {
        List<Quadric> quadsList = new List<Quadric>();

        /// <summary>
        /// Словарь типов переменных
        /// </summary>
        Dictionary<string, Type> types = new Dictionary<string, Type>();
        /// <summary>
        /// Счетчик временных переменных
        /// </summary>
        int tempIndex = 0;
        /// <summary>
        /// Имя временных переменных
        /// </summary>
        public static string TempName = "t";
        /// <summary>
        /// Последнее вычесленное выражение
        /// </summary>
        QExp lastExp;        

        /// <summary>
        /// Флаг обработки присваивания
        /// </summary>
        bool isAssign = false;
        /// <summary>
        /// Переменная возвращающая параметры присваивания
        /// </summary>
        Quadric asgnQuad = null;
        /// <summary>
        /// Счетчик вложенности условных операторов и циклов, для смещения меток внутри блока
        /// </summary>
        int condNest = 0;

        QuadricCollector() { }

        /// <summary>
        /// Основной метод собирающий список команд
        /// </summary>
        /// <param name="root">Узел начала сбора команд</param>
        /// <returns>Список команд трехадресного кода</returns>
        public static List<Quadric> Collect(SyntaxTreeNode root)
        {
            QuadricCollector q = new QuadricCollector();
            root.visit(q);
            return q.quadsList;
        }

        public Type GetType(VarType type)
        {
            string res = "System.";
            switch (type)
            {
                case VarType.Bool: res += "Boolean";
                    break;
                case VarType.Integer: res += "Int32";
                    break;
                case VarType.Double: res += "Double";
                    break;
                default: res += "Object";
                    break;
            }
            return Type.GetType(res);
        }

        public int TypeOrder(Type type)
        {
            switch (type.Name)
            {
                case "Double": return 2;
                case "Int32": return 1;
                case "Boolean": return 0;
                default: return Int32.MaxValue;
            }
        }

        public Type TopType(Type type1, Type type2)
        {

            return TypeOrder(type1) > TypeOrder(type2) ? type1 : type2;
        }

        public Type ResultType(Operation op, Type v1Type, Type v2Type)
        {
            switch (op)
            {
                case Operation.eq:
                case Operation.ge:
                case Operation.gt:
                case Operation.le:
                case Operation.lt: return typeof(Boolean);
                default:
                    if (v1Type.FullName == v2Type.FullName)
                        return v1Type;
                    else
                        return TopType(v1Type, v2Type);
            }
        }

        #region Visitor Methods

        public override void visit(MainBlock _MainBlock)
        {
            _MainBlock.statements.visit(this);
        }

        public override void visit(StatementList _StatementList)
        {
            foreach (Statement s in _StatementList.statements)
                s.visit(this);
        }

        public override void visit(AssignStatement _AssignStatement)
        {
            if (_AssignStatement.To is VarWithIndex)
            {
                VarWithIndex varTo = _AssignStatement.To as VarWithIndex;
                varTo.index.visit(this);                
                QExp v1 = lastExp;
                _AssignStatement.From.visit(this);
                quadsList.Add(new Quadric(QuadOperation.ToIndx, v1, lastExp, new QVar(varTo.Name, types[varTo.Name])));
            }
            else
            {
                isAssign = true;
                _AssignStatement.From.visit(this);
                if (asgnQuad != null)
                {
                    //quadsList.Add(new Quadric(asgnQuad.Operation, asgnQuad.Var1, asgnQuad.Var2, _AssignStatement.To.Name));
                    asgnQuad.Result = new QVar(_AssignStatement.To.Name, types[_AssignStatement.To.Name]);
                    quadsList.Add(asgnQuad);
					asgnQuad = null;
                }
                else
                    quadsList.Add(new Quadric(QuadOperation.Nop, lastExp, null, 
                        new QVar(_AssignStatement.To.Name, types[_AssignStatement.To.Name])));
                isAssign = false;
            }

        }

        public override void visit(VarNode _VarNode)
        {
            lastExp = new QVar(_VarNode.Name, types[_VarNode.Name]);
        }

        public override void visit(VarWithIndex _VarWithIndex)
        {
            _VarWithIndex.index.visit(this);
            quadsList.Add(new Quadric(QuadOperation.FromIndx, new QVar(_VarWithIndex.Name, types[_VarWithIndex.Name]), lastExp,
                /*result*/   (lastExp = new QVar(TempName + ++tempIndex, types[_VarWithIndex.Name])) as QVar));
        }

        public override void visit(IfStatement _IfStatemente)
        {
            _IfStatemente.Condition.visit(this);
            quadsList.Add(new Quadric(QuadOperation.If, lastExp, null, new QConst<int>(quadsList.Count + 2 + condNest)));
            int ind = quadsList.Count;
            condNest++;
            _IfStatemente.ThenBody.visit(this);
            condNest--;
            if (_IfStatemente.ElseBody != null)
            {
                quadsList.Insert(ind, new Quadric(QuadOperation.Goto, null, null, new QConst<int>(quadsList.Count + 2 + condNest)));
                ind = quadsList.Count;
                condNest++;
                _IfStatemente.ElseBody.visit(this);
                condNest--;
                quadsList.Insert(ind, new Quadric(QuadOperation.Goto, null, null, new QConst<int>(quadsList.Count + 1 + condNest)));
            }
            else
                quadsList.Insert(ind, new Quadric(QuadOperation.Goto, null, null, new QConst<int>(quadsList.Count + 1 + condNest)));
        }


        public override void visit(WhileStatement _WhileStatement)
        {
            int ind1 = quadsList.Count;
            _WhileStatement.Condition.visit(this);
            quadsList.Add(new Quadric(QuadOperation.If, lastExp, null, new QConst<int>(quadsList.Count + 2 + condNest)));
            int ind2 = quadsList.Count;
            condNest++;
            _WhileStatement.Body.visit(this);
            condNest--;
            quadsList.Add(new Quadric(QuadOperation.Goto, null, null, new QConst<int>(ind1 + condNest)));
            quadsList.Insert(ind2, new Quadric(QuadOperation.Goto, null, null, new QConst<int>(quadsList.Count + 1 + condNest)));

        }

        public override void visit(Expression _BlockStatement)
        {
            _BlockStatement.visit(this);
        }

        public override void visit(BinaryExpression _BinaryNode)
        {
            if (isAssign)
            {                
                _BinaryNode.Left.visit(this);
                QExp v1 = lastExp;
                _BinaryNode.Right.visit(this);
                asgnQuad = new Quadric((QuadOperation)_BinaryNode.Op, v1, lastExp, new QVar("", null));
            }
            else
            {
                _BinaryNode.Left.visit(this);
                QExp v1 = lastExp;
                _BinaryNode.Right.visit(this);
                quadsList.Add(new Quadric((QuadOperation)_BinaryNode.Op, v1, lastExp,
                    /*result*/   (lastExp = new QVar(TempName + ++tempIndex, 
                                    ResultType(_BinaryNode.Op, v1.Type, lastExp.Type)) as QVar)));
            }

        }

        public override void visit(UnaryExpression _UnaryNode)
        {
            if (isAssign)
            {
                _UnaryNode.Expr.visit(this);
                asgnQuad = new Quadric((QuadOperation)_UnaryNode.Op, lastExp, null, new QVar("", null));
            }
            else
            {
                _UnaryNode.Expr.visit(this);
                quadsList.Add(new Quadric((QuadOperation)_UnaryNode.Op, lastExp, null,
                    /*result*/   (lastExp = new QVar(TempName + ++tempIndex, lastExp.Type)) as QVar));
            }
        }

        public override void visit(BoolConstExpression _BoolConstNode)
        {
            lastExp = new QConst<Boolean>(_BoolConstNode.value);           
        }
        public override void visit(IntConstExpression _IntConstNode)
        {
            lastExp = new QConst<Int32>(_IntConstNode.value);
        }
        public override void visit(DoubleConstExpression _DoubleConstNode)
        {
            lastExp = new QConst<Double>(_DoubleConstNode.value);         
        }

        public override void visit(VarDefStatement _VarDefStatement)
        {
            foreach (VarNode v in _VarDefStatement.Names.Vars)
                types.Add(v.Name, GetType(_VarDefStatement.Type));
        }

        public override void visit(VarWithIndexDefStatement _VarWithIndexDefStatement)
        {
            foreach (VarNode v in _VarWithIndexDefStatement.Vars.Vars)
                types.Add(v.Name, GetType(_VarWithIndexDefStatement.Type));
        }

        #endregion

    }
}
