﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace optimusParser
{
    /// <summary>
    /// Создает остовное дерево графа потока управления.
    /// </summary>
    /// 
    abstract class SpanningTree
    {
        private static List<string> used;  //список раскрытых узлов 
        private static int cnt = 1;  //счётчик для новой нумерации узлов
        private static BaseBlockGraph bg, frame;

        /// <summary>
        /// Возвращает остовное дерево графа потока управления типа BaseBlock.
        /// </summary>    
        
        public static BaseBlockGraph Create(BaseBlockGraph bbg)
        {
            used = new List<string>();
            bg = bbg;
            frame = new BaseBlockGraph();  //создание нового графа-дерева
            AddToFrame(bg.FirstBlocks, null, bg.FirstBlocks.Name);
            return frame;
            
        }

        /* рекурсивная процедура поиска в глубину, по графу потока управления bg строится новое дерево frame
        bb – имя добавляемой вершины в новом дереве
        par – родительский узел в дереве
        old_name – имя узла в графе bg */

        private static void AddToFrame(BaseBlock bb, BaseBlock par, string old_name)
        {
            //добавляем имя узла в графе bg в список раскрытых вершин
            used.Add(old_name);
            //добавляем блок в дерево 
            frame.AddBBlock(bb);
            //если данный блок – не root,
            //то добавляем в дерево связь между bb и его предком par
            if(par != null)
              frame.AddEdge(par.Name, bb.Name);
            List<BaseBlock> childs = bg.GetChildNodes(old_name);
            foreach (BaseBlock bbl in childs)
            {
                //если блок ещё не раскрыт, то
                if (!used.Contains(bbl.Name))
                {
                    //формируем блок дерева с новым именем
                    BaseBlock new_b = new BBlock(cnt.ToString(), bbl.Quads);
                    ++cnt;
                    //рекурсивно запускаем процедуру для всех дочерних вершин
                    AddToFrame(new_b, bb, bbl.Name);
                }
            }
        }
    }
}
